import time
from concurrent import futures

import pytest

from redis_tasks.exceptions import TaskDoesNotExist
from redis_tasks.queue import Queue
from redis_tasks.registries import queue_registry
from redis_tasks.task import Task
from tests.utils import TaskFactory, WorkerFactory, id_list


def test_queue_basics(assert_atomic):
    worker = WorkerFactory()
    q = Queue()
    assert q.name == 'default'
    assert str(q) == '<Queue default>'
    assert repr(q) == "Queue('default')"

    assert q.count() == 0
    with assert_atomic():
        task = q.enqueue_call()
    q.enqueue_call()
    assert q.count() == 2

    assert q.dequeue(worker).id == task.id
    assert q.count() == 1


def test_remove_and_delete(assert_atomic, connection):
    q = Queue()
    q.enqueue_call()
    task = q.enqueue_call()
    assert q.count() == 2
    assert connection.exists(task.key)
    with assert_atomic():
        q.remove_and_delete(task)
    assert not connection.exists(task.key)
    assert q.count() == 1
    with pytest.raises(TaskDoesNotExist):
        q.remove_and_delete(task)
    assert q.count() == 1


def test_get_tasks():
    q = Queue()
    tasks = [TaskFactory() for i in range(3)]
    for task in tasks:
        task._save()
        q.push(task)
    assert q.get_task_ids() == id_list(tasks)
    assert id_list(q.get_tasks()) == id_list(tasks)
    q.remove_and_delete(tasks[1])
    assert q.get_task_ids() == [tasks[0].id, tasks[2].id]


def test_queue_all():
    q1 = Queue('a')
    q2 = Queue('b')
    q3 = Queue('c')
    q2.enqueue_call()
    q3.enqueue_call()
    assert [x.name for x in Queue.all()] == ['b', 'c']
    q2.delete()
    q3.empty()
    q1.enqueue_call()
    assert [x.name for x in Queue.all()] == ['a', 'c']


def test_unblocking(connection):
    worker = WorkerFactory()
    q = Queue()
    q.enqueue_call()
    assert connection.llen(q.unblock_key)
    assert q.dequeue(worker) is not None
    assert connection.llen(q.unblock_key)
    assert q.dequeue(worker) is None
    assert not connection.llen(q.unblock_key)


def test_push(assert_atomic):
    q = Queue()
    q.enqueue_call()
    q.enqueue_call()
    task = TaskFactory()
    with assert_atomic():
        q.push(task)
    assert q.get_task_ids()[-1] == task.id

    task = TaskFactory()
    q.push(task, at_front=True)
    assert q.get_task_ids()[0] == task.id


class TestEmptyAndDelete:
    def test_empty_empty_queue(self):
        q = Queue()
        q.empty()

    def test_empty(self, assert_atomic, connection, mocker):
        q = Queue()
        t1 = q.enqueue_call()
        t2 = q.enqueue_call()
        assert connection.exists(t1.key)
        with assert_atomic():
            q.empty()
        assert not connection.exists(q.key)
        assert not connection.exists(q.unblock_key)
        assert q.name in queue_registry.get_names()
        assert not connection.exists(t1.key)
        assert not connection.exists(t2.key)

    def test_delete(self, assert_atomic, connection, mocker):
        q = Queue()
        q.enqueue_call()
        q.enqueue_call()
        with assert_atomic():
            q.delete()
        assert not connection.exists(q.key)
        assert not connection.exists(q.unblock_key)
        assert q.name not in queue_registry.get_names()

    def test_delete_transaction(self, assert_atomic, connection, mocker):
        q = Queue()
        task1 = q.enqueue_call()
        task2 = None

        def task_delete(task_ids, **kwargs):
            nonlocal task2
            if not task2:
                # interrupt the transaction with the creation of a new task
                task2 = q.enqueue_call()
            else:
                # assert that the previous transaction attempt was canceled
                assert connection.exists(task1.key)
            orig_delete_many(task_ids, **kwargs)

        orig_delete_many = Task.delete_many
        mocker.patch.object(Task, 'delete_many', new=task_delete)
        assert connection.exists(task1.key)
        q.empty()
        assert not connection.exists(task1.key)
        assert not connection.exists(task2.key)


def test_dequeue(connection):
    worker = WorkerFactory()
    q = Queue()
    assert q.dequeue(worker) is None

    task1 = q.enqueue_call()
    task2 = q.enqueue_call()
    assert connection.llen(q.unblock_key) == 2

    assert q.dequeue(worker).id == task1.id
    assert worker.current_task_id == task1.id
    assert connection.lrange(worker.task_key, 0, -1) == [task1.id.encode()]

    assert q.dequeue(worker).id == task2.id
    assert worker.current_task_id == task2.id
    assert connection.lrange(worker.task_key, 0, -1) == [task2.id.encode(), task1.id.encode()]
    assert connection.llen(q.unblock_key) == 2

    # Clear block_key when queue is empty
    assert q.dequeue(worker) is None
    assert connection.llen(q.unblock_key) == 0


def test_await_multi():
    q1 = Queue('first')
    q2 = Queue('second')

    blocked = False

    def in_thread():
        nonlocal blocked
        blocked = True
        q = Queue.await_multi([q1, q2], 1)
        blocked = False
        return q

    with futures.ThreadPoolExecutor(max_workers=5) as executor:
        future = executor.submit(in_thread)
        time.sleep(0.005)
        assert blocked
        q1.push(TaskFactory())
        time.sleep(0.005)
        assert not blocked
        assert future.result().name == q1.name

        future = executor.submit(in_thread)
        time.sleep(0.005)
        q2.push(TaskFactory())
        q1.push(TaskFactory())
        assert future.result().name == q2.name

    assert Queue.await_multi([q1, q2], 1).name == q1.name


@pytest.mark.slow
def test_queue_await_multi_empty():
    assert Queue.await_multi([Queue()], 1) is None



from redis_tasks.queue import Queue
from redis_tasks.task import Task
from redis_tasks.worker import Worker

worker_sequence = 0
queue_sequence = 0
task_sequence = 0


def _stub():
    pass


def TaskFactory():
    global task_sequence
    task_sequence += 1
    task = Task(func=_stub)
    task.id = f'task_{task_sequence}'
    return task


def WorkerFactory(*, queues=None):
    global worker_sequence
    worker_sequence += 1
    id = f'worker_{worker_sequence}'
    description = f"Worker {worker_sequence}"
    if queues is None:
        queues = [QueueFactory()]
    return Worker(id, description=description, queues=queues)


def QueueFactory():
    global queue_sequence
    queue_sequence += 1
    name = f"queue_{queue_sequence}"
    return Queue(name)


def id_list(lst):
    return [x.id for x in lst]


class AnythingType(object):
    def __eq__(self, other):
        return True

    def __ne__(self, other):
        return False

    def __repr__(self):
        return 'Anything'


Anything = AnythingType()


class SomethingType(object):
    def __eq__(self, other):
        return other is not None

    def __ne__(self, other):
        return other is None

    def __repr__(self):
        return 'Something'


Something = SomethingType()



import copy
import datetime
import multiprocessing
import os
import signal
from contextlib import contextmanager
from unittest import mock

import pytest

from redis_tasks import conf

os.environ[conf.ENVIRONMENT_VARIABLE] = 'tests.app.settings'


def pytest_addoption(parser):
    parser.addoption("--run-slow", action="store_true", help="run slow tests")


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    if not item.config.getoption("--run-slow"):
        if item.keywords.get('slow'):
            pytest.skip("Test is marked as slow")


def pytest_unconfigure(config):
    do_clear_redis()


class ModifiableSettings:
    pass


@pytest.fixture()
def settings():
    from redis_tasks import conf

    # ensure the settings are initialized
    conf.settings.DEFAULT_TASK_TIMEOUT
    original_dict = conf.settings.__dict__
    new_dict = copy.deepcopy(original_dict)
    conf.settings.__dict__ = new_dict
    mod_settings = ModifiableSettings()
    mod_settings.__dict__ = new_dict
    yield mod_settings
    conf.settings.__dict__ = original_dict


@pytest.fixture(autouse=True, scope="function")
def mock_signal(mocker):
    """Prevent tests from messing with the signal handling of the pytest process"""
    main_pid = os.getpid()
    orig_signal = signal.signal

    def wrapped_signal(*args, **kwargs):
        if os.getpid() == main_pid:
            return
        else:
            return orig_signal(*args, **kwargs)

    mocker.patch('signal.signal', new=wrapped_signal)
    yield


@pytest.fixture(autouse=True, scope="function")
def clear_redis():
    do_clear_redis()
    yield


def do_clear_redis():
    from redis_tasks import conf
    with conf.connection.pipeline() as pipeline:
        for key in conf.connection.scan_iter(conf.construct_redis_key('*')):
            pipeline.delete(key)
        pipeline.execute()


@pytest.fixture(scope="function")
def connection():
    from redis_tasks import conf
    yield conf.connection


class AtomicRedis:
    def __init__(self, wrap, exceptions):
        self.wrapped = wrap
        self.exceptions = exceptions + ['ftime']
        self._atomic_counter = 0

    def __getattr__(self, name):
        __tracebackhide__ = True
        if name in self.exceptions:
            return getattr(self.wrapped, name)
        if name not in ["pipeline", "transaction", "register_script"]:
            raise Exception(f"Attempted call to connection.{name} in assert_atomic")
        if self._atomic_counter > 0:
            raise Exception("Second call to connection function in assert_atomic")
        self._atomic_counter += 1
        return getattr(self.wrapped, name)


@pytest.fixture(scope="function")
def assert_atomic(mocker):
    from redis_tasks import conf

    @contextmanager
    def cm(*, exceptions=[]):
        real_connection = conf.connection._wrapped
        with mock.patch.dict(conf.connection.__dict__,
                             _wrapped=AtomicRedis(real_connection, exceptions)):
            yield
    yield cm


class TimeMocker:
    def __init__(self, mocker, target):
        from redis_tasks.utils import utcnow
        self.seq = 0
        self.mocker = mocker
        self.target = target
        self.now = utcnow().replace(microsecond=0)
        self.step()

    def step(self, seconds=1):
        self.seq += 1
        self.now += datetime.timedelta(seconds=seconds)
        self.mocker.patch(self.target, return_value=self.now)
        return self.now


@pytest.fixture()
def time_mocker(mocker):
    yield lambda target: TimeMocker(mocker, target)


@pytest.fixture(autouse=True)
def kill_child_processes():
    yield
    for child in multiprocessing.active_children():
        print(f"Killing leftover process {child.name}")
        os.kill(child.pid, signal.SIGKILL)


def _stub(*args, **kwargs):
    return _stub.mock(*args, **kwargs)


@pytest.fixture
def stub():
    _stub.mock = mock.Mock(return_value=None)
    _stub.path = f"{_stub.__module__}._stub"
    return _stub


@pytest.fixture(autouse=True)
def _task_stub(stub, mocker):
    from redis_tasks import Task

    def my_init(self, *args, **kwargs):
        __tracebackhide__ = True
        if not args and not kwargs:
            args = [stub]
        return orig_init(self, *args, **kwargs)

    orig_init = Task.__init__
    mocker.patch.object(Task, '__init__', my_init)
    yield



import datetime

import pytz

from redis_tasks.smear_dst import DstSmearingTz


def test_utc():
    tz = DstSmearingTz('UTC')
    dt = datetime.datetime(2000, 1, 1, 12, 1, 5)
    assert tz.to_utc(dt) == dt.replace(tzinfo=pytz.UTC)
    assert tz.from_utc(dt.replace(tzinfo=pytz.UTC)) == dt


def test_japan():
    tz = DstSmearingTz('Asia/Tokyo')
    dt = datetime.datetime(2000, 1, 1, 12, 1, 5)
    assert tz.from_utc(tz.to_utc(dt)) == dt


def test_germany():
    tz = DstSmearingTz('Europe/Berlin')

    dt = datetime.datetime(2017, 1, 1, 18, 30, 23)
    assert tz.to_utc(dt) == datetime.datetime(2017, 1, 1, 17, 30, 23, tzinfo=pytz.UTC)
    assert tz.from_utc(tz.to_utc(dt)) == dt

    dt = datetime.datetime(2017, 3, 26, 2, 17, 30)
    assert tz.to_utc(dt) == datetime.datetime(2017, 3, 26, 0, 51, 40, tzinfo=pytz.UTC)
    assert tz.from_utc(tz.to_utc(dt)) == dt

    dt = datetime.datetime(2017, 10, 29, 2, 30, 0)
    assert tz.to_utc(dt) == datetime.datetime(2017, 10, 29, 1, 0, 0, tzinfo=pytz.UTC)
    assert tz.from_utc(tz.to_utc(dt)) == dt



import time

import pytest

from redis_tasks.worker_process import Maintenance


@pytest.mark.slow
def test_run_if_necessary(mocker, settings):
    settings.MAINTENANCE_FREQ = 2
    maintenance = Maintenance()
    run = mocker.patch.object(Maintenance, 'run')
    maintenance.run_if_neccessary()
    assert run.call_count == 1

    maintenance.run_if_neccessary()
    assert run.call_count == 1

    time.sleep(2)
    maintenance.run_if_neccessary()
    assert run.call_count == 2



import datetime
import logging
import math
import sys
import time
from functools import partial
from operator import attrgetter

import click
from redis.exceptions import ConnectionError

from redis_tasks import __version__, scheduler_main, worker_main
from redis_tasks.queue import Queue
from redis_tasks.worker import Worker, WorkerState

red = partial(click.style, fg='red')
green = partial(click.style, fg='green')
yellow = partial(click.style, fg='yellow')


@click.group()
@click.version_option(__version__)
@click.option(
    '--pythonpath', default='',
    help='A directory to add to the Python path, defaults to the work directory')
def main(pythonpath):
    """redis_tasks command line tool"""
    sys.path.insert(0, pythonpath)


@main.command()
@click.option('--all', is_flag=True, help='Empty all queues')
@click.option('--delete', is_flag=True, help='Completely remove queues')
@click.argument('queues', nargs=-1)
def empty(all, delete, queues, **options):
    """Empty selected queues"""

    if all:
        queues = Queue.all()
    elif queues:
        queues = [Queue(q) for q in queues]
    else:
        raise click.UsageError("No queues specified")

    if not queues:
        click.echo('Nothing to do')
        sys.exit(0)

    for queue in queues:
        if delete:
            queue.delete()
        else:
            queue.empty()
        click.echo(f'Queue f{queue.name} emptied')


def configure_logging(verbose, quiet, **options):
    if verbose and quiet:
        raise click.UsageError("flags --verbose and --quiet are mutually exclusive")

    level = (verbose and 'DEBUG') or (quiet and 'WARNING') or 'INFO'
    logging.basicConfig(level=level, style='{', datefmt='%H:%M:%S',
                        format='{asctime} {levelname:5.5} {name:>20}: {message}')


@main.command()
@click.option('--verbose', '-v', is_flag=True, help='Show more output')
@click.option('--quiet', '-q', is_flag=True, help='Show less output')
def scheduler(**options):
    """Run the redis_tasks scheduler"""
    configure_logging(**options)
    try:
        scheduler_main()
    except ConnectionError as e:
        print(e)
        sys.exit(1)


@main.command()
@click.option('--burst', '-b', is_flag=True, help='Run in burst mode (quit after all work is done)')
@click.option('--description', '-d', help='Specify a different description')
@click.option('--verbose', '-v', is_flag=True, help='Show more output')
@click.option('--quiet', '-q', is_flag=True, help='Show less output')
@click.argument('queues', nargs=-1)
def worker(burst, description, queues, **options):
    """Run a redis_tasks worker"""
    configure_logging(**options)
    queues = queues or ['default']
    try:
        worker_main(queues, description=description, burst=burst)
    except ConnectionError as e:
        print(e)
        sys.exit(1)


@main.command()
@click.option('--interval', '-i', type=float, help='Updates stats every N seconds')
@click.option('--by-queue', is_flag=True, help='Shows workers by queue')
@click.argument('queues', nargs=-1)
def info(interval, by_queue, queues, **options):
    """Display information about active queues and workers"""
    if queues:
        queues = [Queue(q) for q in queues]

    def display():
        show_queues(queues)
        click.echo('')
        show_workers(queues, by_queue)
        click.echo('')
    try:
        if interval:
            while True:
                click.clear()
                display()
                click.echo('Updated: %s' % datetime.datetime.now())
                time.sleep(interval)
        else:
            display()
    except ConnectionError as e:
        click.echo(e)
        sys.exit(1)
    except KeyboardInterrupt:
        click.echo()
        sys.exit(0)


def print_separator():
    click.echo("=" * (min(40, click.get_terminal_size()[0])))


def show_queues(queues):
    if not queues:
        queues = Queue.all()

    counts = {q: q.count() for q in queues}

    chart_width = 20
    chart_max = max((1, *counts.values()))
    # Round up to the next fixed marker or the next power of two times 1000
    chart_max = next((x for x in [20, 50, 100, 200, 400, 600, 800, 1000]
                      if x >= chart_max),
                     2 ** math.ceil(math.log2(chart_max / 1000)) * 1000)

    click.secho("Queues", bold=True)
    print_separator()
    for q in sorted(queues, key=attrgetter('name')):
        count = counts[q]
        chart = green('|' + '█' * int(count / chart_max * chart_width))
        click.echo(f'{q.name:<12} {chart} {count}')

    click.echo(f'{len(queues)} queue(s), {sum(counts.values())} task(s) total')


def show_workers(queues, by_queue):
    def state_symbol(state):
        return {
            WorkerState.BUSY: red('busy'),
            WorkerState.IDLE: green('idle')
        }.get(state, state)

    workers = Worker.all()
    all_queues = set(Queue.all()) | {q for w in workers for q in w.queues}
    if queues:
        workers = [w for w in workers
                   if any(set(w.queues) & set(queues))]
    else:
        queues = all_queues

    if not by_queue:
        click.secho("Workers", bold=True)
    else:
        click.secho("Workers per Queue", bold=True)
    print_separator()
    if not by_queue:
        for worker in sorted(workers, key=attrgetter('description')):
            worker_queues = ', '.join(q.name for q in worker.queues)
            click.echo(f'{worker.description} {state_symbol(worker.state)}: {worker_queues}')
    else:
        queue_map = {q: [] for q in all_queues}
        for w in workers:
            for q in w.queues:
                queue_map[q].append(w)
        max_qname = max(len(q.name) for q in queues)
        for queue in sorted(queues, key=attrgetter('name')):
            q_workers = queue_map[queue]
            workers_str = ", ".join(sorted(f'{w.description} {state_symbol(w.state)}'
                                           for w in q_workers))
            if not workers_str:
                workers_str = '–'
            click.echo(f'{queue.name:>{max_qname}}: {workers_str}')

    click.echo(f'{len(workers)} worker(s)')


if __name__ == '__main__':
    main()



import datetime
import time
from concurrent import futures

import pytest
import pytz

from redis_tasks import Queue
from redis_tasks.scheduler import (
    CrontabSchedule, Mutex, PeriodicSchedule, Scheduler, SchedulerEntry)
from redis_tasks.task import TaskOutcome
from redis_tasks.utils import one, utcnow
from tests.utils import WorkerFactory


def iter_schedule(schedule, start, count):
    for i in range(count):
        start = schedule.get_next(start)
        yield start


def test_crontab_schedule(settings):
    settings.TIMEZONE = "Europe/Berlin"
    tz = pytz.timezone('Europe/Berlin')
    schedule = CrontabSchedule('30 2 * * *')
    start = datetime.datetime(2017, 3, 24, tzinfo=pytz.utc)
    assert list(dt.astimezone(tz) for dt in iter_schedule(schedule, start, 5)) == [
        tz.localize(dt, is_dst=None) for dt in [
            datetime.datetime(2017, 3, 24, 2, 30),
            datetime.datetime(2017, 3, 25, 2, 30),
            # DST Transition
            datetime.datetime(2017, 3, 26, 3, 0),
            datetime.datetime(2017, 3, 27, 2, 30),
            datetime.datetime(2017, 3, 28, 2, 30),
        ]]

    start = datetime.datetime(2017, 10, 27, tzinfo=pytz.utc)
    assert list(dt.astimezone(tz) for dt in iter_schedule(schedule, start, 5)) == [
        tz.localize(datetime.datetime(2017, 10, 27, 2, 30)),
        tz.localize(datetime.datetime(2017, 10, 28, 2, 30), is_dst=True),
        tz.localize(datetime.datetime(2017, 10, 29, 2, 0), is_dst=False),
        tz.localize(datetime.datetime(2017, 10, 30, 2, 30), is_dst=False),
        tz.localize(datetime.datetime(2017, 10, 31, 2, 30)),
    ]


def test_periodic_schedule(settings):
    settings.TIMEZONE = "Europe/Berlin"
    tz = pytz.timezone('Europe/Berlin')

    # Test transition to DST
    schedule = PeriodicSchedule(minutes=90, start_at="01:00")
    start = tz.localize(datetime.datetime(2017, 3, 26))
    assert list(dt.astimezone(tz) for dt in iter_schedule(schedule, start, 6)) == [
        tz.localize(dt, is_dst=None) for dt in [
            datetime.datetime(2017, 3, 26, 1, 0),
            datetime.datetime(2017, 3, 26, 3, 30),
            datetime.datetime(2017, 3, 26, 5, 0),
            datetime.datetime(2017, 3, 26, 6, 30),
            datetime.datetime(2017, 3, 26, 8, 0),
            datetime.datetime(2017, 3, 26, 9, 30),
        ]]

    # Test night-wrap
    schedule = PeriodicSchedule(minutes=17, start_at="23:00")
    start = tz.localize(datetime.datetime(2017, 4, 1, 22))
    assert list(dt.astimezone(tz) for dt in iter_schedule(schedule, start, 6)) == [
        tz.localize(dt, is_dst=None) for dt in [
            datetime.datetime(2017, 4, 1, 23, 00),
            datetime.datetime(2017, 4, 1, 23, 17),
            datetime.datetime(2017, 4, 1, 23, 34),
            datetime.datetime(2017, 4, 1, 23, 51),
            datetime.datetime(2017, 4, 2, 23, 00),
            datetime.datetime(2017, 4, 2, 23, 17),
        ]]

    assert (schedule.get_next(tz.localize(datetime.datetime(2017, 4, 1, 23, 10))) ==
            tz.localize(datetime.datetime(2017, 4, 1, 23, 17)))

    # Test transition from DST
    schedule = PeriodicSchedule(minutes=40, start_at="01:00")
    start = tz.localize(datetime.datetime(2017, 10, 29, 0))
    assert list(dt.astimezone(tz) for dt in iter_schedule(schedule, start, 6)) == [
        tz.localize(datetime.datetime(2017, 10, 29, 1, 00)),
        tz.localize(datetime.datetime(2017, 10, 29, 1, 40), is_dst=True),
        tz.localize(datetime.datetime(2017, 10, 29, 2, 20), is_dst=True),
        tz.localize(datetime.datetime(2017, 10, 29, 2, 00), is_dst=False),
        tz.localize(datetime.datetime(2017, 10, 29, 2, 40), is_dst=False),
        tz.localize(datetime.datetime(2017, 10, 29, 3, 20)),
    ]


class TestSchedulerEntry:
    def time(self, *args):
        return datetime.datetime(2000, 1, 1, *args, tzinfo=pytz.utc)

    def test_sanity_check(self):
        schedule = CrontabSchedule('* * * * *')
        with pytest.raises(ValueError):
            SchedulerEntry("a", {"task": "broken", "schedule": schedule})

    def test_init_and_save(self, time_mocker, connection, settings,
                           assert_atomic, mocker, stub):
        time = time_mocker('redis_tasks.scheduler.utcnow')
        schedule = CrontabSchedule('* * * * *')
        se = SchedulerEntry("a", {"task": stub.path, "schedule": schedule})
        assert se.id == "a"
        assert se.prev_run == time.now
        assert se.next_run and se.next_run != se.prev_run
        assert se.prev_task_id is None
        time.step()
        with assert_atomic():
            se.save()
        assert se.last_save == time.now

        time.step()
        loaded = SchedulerEntry("a", {"task": stub.path, "schedule": schedule})
        assert loaded.prev_run == se.prev_run
        assert loaded.next_run == se.next_run
        assert loaded.prev_task_id is None

        se.prev_task_id = "b"
        se.save()
        loaded = SchedulerEntry("a", {"task": stub.path, "schedule": schedule})
        assert loaded.prev_task_id == "b"

        settings.SCHEDULER_MAX_CATCHUP = 60
        se.save()
        assert connection.ttl(se.key) == 24 * 60 * 60
        settings.SCHEDULER_MAX_CATCHUP = 24 * 60 * 60
        se.save()
        assert connection.ttl(se.key) == 24 * 60 * 60 * 5

        assert se.singleton is True
        assert se.queue.name == "default"

        se = SchedulerEntry("a", {
            "task": stub.path,
            "schedule": schedule,
            "singleton": False,
            "queue": "myq",
        })
        assert se.singleton is False
        assert se.queue.name == "myq"

    def test_enqueue(self, assert_atomic, stub):
        schedule = CrontabSchedule('* * * * *')
        queue = Queue()
        se = SchedulerEntry("a", {"task": stub.path, "schedule": schedule})
        with assert_atomic():
            task = se.enqueue()
        assert task.origin == "default"
        assert task._get_func() == stub

        se = SchedulerEntry("a", {"task": stub.path, "schedule": schedule,
                                  "args": ["a"], "kwargs": {"c": "d"}})

        task = se.enqueue()
        assert task.args == ["a"]
        assert task.kwargs == {"c": "d"}
        assert queue.count() == 2

    def test_is_enqueued(self, stub):
        schedule = CrontabSchedule('* * * * *')
        queue = Queue()
        worker = WorkerFactory()
        worker.startup()
        se = SchedulerEntry("a", {"task": stub.path, "schedule": schedule})
        assert se.is_enqueued() is False

        task = se.enqueue()
        assert se.is_enqueued() is True
        task.cancel()
        assert se.is_enqueued() is False

        se.enqueue()
        assert se.is_enqueued() is True
        task = queue.dequeue(worker)
        assert se.is_enqueued() is True
        worker.start_task(task)
        assert se.is_enqueued() is True
        worker.end_task(task, TaskOutcome("success"))
        assert se.is_enqueued() is False

        se.enqueue()
        task = queue.dequeue(worker)
        worker.start_task(task)
        assert se.is_enqueued() is True
        worker.end_task(task, TaskOutcome("failure"))
        assert se.is_enqueued() is False

    def test_process_singleton(self, mocker, settings, stub):
        queue = Queue()
        settings.SCHEDULER_MAX_CATCHUP = 60 * 60 * 24
        schedule = mocker.Mock()
        get_next = schedule.get_next
        get_next.return_value = self.time(1)
        se = SchedulerEntry("a", {"task": stub.path, "schedule": schedule,
                                  "singleton": True})
        se.prev_run = self.time(0)

        get_next.reset_mock()
        se.process(self.time(0))
        assert se.next_run == self.time(1)
        get_next.assert_called_once_with(self.time(0))
        assert se.is_enqueued() is False

        se.process(self.time(1))
        assert se.is_enqueued() is True
        se.process(self.time(2))
        assert queue.count() == 1
        queue.empty()

        assert se.is_enqueued() is False
        se.process(self.time(2))
        assert se.is_enqueued() is True

        queue.empty()
        get_next.reset_mock()
        get_next.side_effect = [self.time(1), self.time(3)]
        se.prev_run = self.time(0)
        se.process(self.time(2))
        assert queue.count() == 1
        assert get_next.call_count == 2
        get_next.assert_any_call(self.time(0))
        get_next.assert_called_with(self.time(2))
        assert se.next_run == self.time(3)

    def test_process_max_catchup(self, mocker, settings, stub):
        settings.SCHEDULER_MAX_CATCHUP = 60 * 60 * 1
        schedule = mocker.Mock()
        get_next = schedule.get_next
        get_next.return_value = self.time(4)
        se = SchedulerEntry("a", {"task": stub.path, "schedule": schedule,
                                  "singleton": True})
        get_next.reset_mock()
        se.prev_run = self.time(0)
        se.process(self.time(2))
        get_next.assert_called_once_with(self.time(1))
        assert se.is_enqueued() is False

    def test_process_non_singleton(self, mocker, settings, stub):
        queue = Queue()
        settings.SCHEDULER_MAX_CATCHUP = 60 * 60 * 24
        schedule = mocker.Mock()
        get_next = schedule.get_next
        get_next.return_value = self.time(2)
        se = SchedulerEntry("a", {"task": stub.path, "schedule": schedule,
                                  "singleton": False})

        get_next.reset_mock()
        se.prev_run = self.time(0)
        se.process(self.time(1))
        get_next.assert_called_once_with(self.time(0))
        assert se.next_run == self.time(2)
        assert se.is_enqueued() is False
        assert se.prev_run == self.time(0)

        get_next.reset_mock()
        get_next.side_effect = [self.time(2), self.time(3)]
        se.process(self.time(2))
        assert se.is_enqueued() is True
        assert get_next.call_count == 2
        get_next.assert_any_call(self.time(0))
        get_next.assert_any_call(self.time(2))
        assert se.prev_run == self.time(2)

        queue.empty()
        get_next.reset_mock()
        get_next.side_effect = [self.time(3), self.time(4), self.time(5)]
        se.process(self.time(4))
        assert se.next_run == self.time(5)
        assert queue.count() == 2
        assert get_next.call_count == 3


class TestMutex:
    def test_basic(self, caplog):
        with Mutex(timeout=1) as m:
            m.extend()
        assert not caplog.records

    @pytest.mark.slow
    def test_blocked(self, caplog):
        with Mutex(timeout=3):
            with pytest.raises(RuntimeError):
                with Mutex(timeout=0.1):
                    pass
        assert 'already running' in one(caplog.records).message

    @pytest.mark.slow
    def test_unblocked(self, caplog):
        with Mutex(timeout=0.01):
            with Mutex(timeout=0.01):
                assert 'already running' in one(caplog.records).message

    @pytest.mark.slow
    def test_extend(self):
        with Mutex(timeout=0.02) as m:
            time.sleep(0.01)
            m.extend()
            time.sleep(0.01)
            m.extend()
            time.sleep(0.01)
            m.extend()

    def test_missed_extend(self, caplog):
        with Mutex(timeout=0.001) as m:
            time.sleep(0.002)
            with pytest.raises(RuntimeError):
                m.extend()


class FixedSchedule:
    def __init__(self, times):
        self.times = times

    def get_next(self, after):
        return next(t for t in self.times if t > after)


class TestScheduler:
    def test_basic(self, settings, mocker, time_mocker, stub):
        mocktime = time_mocker('redis_tasks.scheduler.utcnow')
        schedule = mocker.Mock()
        schedule.get_next.return_value = mocktime.now
        settings.SCHEDULE = {'mytask': {
            'task': stub.path,
            'schedule': schedule,
        }}
        queue = Queue()

        def stop_scheduler():
            time.sleep(0.02)
            scheduler.shutdown_requested.set()

        scheduler = Scheduler()
        with futures.ThreadPoolExecutor(max_workers=2) as executor:
            executor.submit(stop_scheduler)
            scheduler.run()
        assert queue.count() == 1

    def test_scheduling(self, settings, mocker, stub):
        in_ms = lambda ms: utcnow() + datetime.timedelta(milliseconds=ms)  # noqa
        settings.SCHEDULE = {
            'a': {'task': stub.path,
                  'args': ["a"],
                  'singleton': False,
                  'schedule': FixedSchedule([in_ms(10), in_ms(11), in_ms(80), in_ms(1000)])},
            'b': {'task': stub.path,
                  'args': ["b"],
                  'singleton': False,
                  'schedule': FixedSchedule([in_ms(40), in_ms(1000)])},
        }
        queue = Queue()

        def stop_scheduler():
            time.sleep(0.1)
            scheduler.shutdown_requested.set()

        scheduler = Scheduler()
        with futures.ThreadPoolExecutor(max_workers=2) as executor:
            executor.submit(stop_scheduler)
            scheduler.run()
        assert [t.args for t in queue.get_tasks()] == [["a"], ["a"], ["b"], ["a"]]



import random

import pytest

from redis_tasks import Queue, TWorker
from redis_tasks.contrib.graph import GraphMiddleware, TaskGraph, chain


@pytest.fixture
def with_mw(mocker):
    mocker.patch('redis_tasks.conf.task_middleware', new=[GraphMiddleware])
    mocker.patch('redis_tasks.task.task_middleware', new=[GraphMiddleware])


def test_middleware_check(mocker):
    with pytest.raises(Exception):
        TaskGraph().enqueue()

    mocker.patch('redis_tasks.conf.task_middleware', new=[GraphMiddleware])
    TaskGraph().enqueue()


def test_chain(stub, with_mw, mocker):
    def queue_assert(*args):
        assert Queue().count() == 0

    stub.mock.side_effect = queue_assert
    chain([
        dict(func=stub, args=["a"]),
        dict(func=stub, args=["b"]),
        dict(func=stub, args=["c"]),
    ]).enqueue()
    assert Queue().count() == 1
    w = TWorker()
    assert w.run() == 3
    assert stub.mock.call_count == 3
    assert stub.mock.call_args_list == [
        mocker.call('a'), mocker.call("b"), mocker.call("c")]


def test_graph(stub, with_mw, mocker):
    def queue_assert(rnd, queue_size):
        assert Queue().count() == queue_size

    stub.mock.side_effect = queue_assert
    graph = TaskGraph()
    # We construct the following graph:
    #  1  2f
    # /|\/|
    # ||/\|
    # |3  4
    # |/
    # 0
    tasks = [dict(func=stub, args=(random.randrange(10000), queue_size))
             for queue_size in [0, 1, 0, 1, 1]]
    nodes = [graph.add_task(t) for t in tasks]
    edges = [(1, 0), (1, 3), (1, 4), (2, 3), (2, 4), (3, 0)]
    for a, b in edges:
        graph.add_dependency(nodes[a], nodes[b])
    graph.enqueue()
    assert Queue().count() == 2
    w = TWorker()
    assert w.run() == 5
    assert stub.mock.call_args_list == [
        mocker.call(*tasks[i]["args"]) for i in [1, 2, 3, 4, 0]]



import click
import pytest

from redis_tasks.queue import Queue
from tests.utils import QueueFactory, WorkerFactory


@pytest.fixture
def cli_run():
    from click.testing import CliRunner

    from redis_tasks.cli import main as cli_main
    runner = CliRunner()

    def run(*args):
        return runner.invoke(cli_main, args,
                             catch_exceptions=False, standalone_mode=False)

    return run


def test_empty(cli_run, stub):
    queues = [QueueFactory() for i in range(5)]
    for q in queues:
        q.enqueue_call(stub)

    with pytest.raises(click.UsageError):
        cli_run('empty')

    assert all(q.count() == 1 for q in queues)

    cli_run('empty', queues[0].name, queues[1].name)
    assert queues[0].count() == 0
    assert queues[1].count() == 0
    assert queues[2].count() == 1

    assert set(queues) == set(Queue.all())
    cli_run('empty', '--delete', queues[1].name, queues[2].name)
    assert set(queues) - {queues[1], queues[2]} == set(Queue.all())

    cli_run('empty', '--all')
    assert all(q.count() == 0 for q in queues)


def test_worker(cli_run, mocker):
    worker_main = mocker.patch('redis_tasks.cli.worker_main')
    log_config = mocker.patch('logging.basicConfig')
    cli_run('worker')
    assert worker_main.called_once_with(['default'])
    assert log_config.called_once()
    assert log_config.call_args[1]["level"] == "INFO"

    cli_run('worker', '--quiet')
    assert log_config.call_args[1]["level"] == "WARNING"

    cli_run('worker', '--verbose')
    assert log_config.call_args[1]["level"] == "DEBUG"

    cli_run('worker', '-d', 'foo')
    assert worker_main.call_args[1]["description"] == 'foo'


def test_info(cli_run, stub):
    cli_run('info')

    for i in range(2):
        QueueFactory().enqueue_call(stub)

    for i in range(3):
        WorkerFactory().startup()

    result = cli_run('info')
    assert "2 queue(s), 2 task(s) total" in result.output
    assert "3 worker(s)" in result.output



import sys
from contextlib import contextmanager

import pytest

from redis_tasks.exceptions import TaskAborted, WorkerShutdown
from redis_tasks.task import Task
from tests.utils import Something


def test_successful_execute(mocker, stub):
    task = Task(stub, ["foo"], {"foo": "bar"})
    outcome = task.execute()
    assert stub.mock.called_once_with("foo", foo="bar")
    assert outcome.outcome == 'success'

    task = Task(stub)
    stub.mock.reset_mock()
    outcome = task.execute()
    assert stub.mock.called_once_with("foo", foo="bar")
    assert outcome.outcome == 'success'


def test_failed_execute(mocker, stub):
    stub.mock.side_effect = ValueError("TestException")
    task = Task(stub)
    outcome = task.execute()
    assert stub.mock.called_once_with()
    assert outcome.outcome == 'failure'
    assert outcome.message.splitlines()[-1] == 'ValueError: TestException'


def test_aborted_execute(mocker, stub):
    stub.mock.side_effect = WorkerShutdown()
    task = Task(stub)
    outcome = task.execute()
    assert stub.mock.called_once_with()
    assert outcome.outcome == 'failure'
    assert outcome.message.splitlines()[-1] == 'redis_tasks.exceptions.TaskAborted: Worker shutdown'


def test_broken_task(stub):
    task = Task(stub)
    task.func_name = "nonimportable.function"
    outcome = task.execute()
    assert outcome.outcome == 'failure'
    assert outcome.message.splitlines()[-1].startswith(
        'RuntimeError: Failed to import task function')


def test_shutdown_cm(mocker, stub):
    @contextmanager
    def entry_shutdown_cm():
        raise WorkerShutdown()
        yield

    task = Task(stub)
    outcome = task.execute(shutdown_cm=entry_shutdown_cm())
    assert not stub.mock.called
    assert outcome.outcome == 'failure'
    assert 'Worker shutdown' in outcome.message.splitlines()[-1]

    @contextmanager
    def exit_shutdown_cm():
        yield
        raise WorkerShutdown()

    stub.mock.reset_mock()
    outcome = task.execute(shutdown_cm=exit_shutdown_cm())
    assert stub.mock.called_once_with()
    assert outcome.outcome == 'failure'
    assert 'Worker shutdown' in outcome.message.splitlines()[-1]

    in_cm = False

    @contextmanager
    def reporting_cm():
        nonlocal in_cm
        in_cm = True
        yield
        in_cm = False

    def checking_func():
        assert in_cm

    stub.mock.side_effect = checking_func
    outcome = task.execute(shutdown_cm=reporting_cm())
    assert outcome.outcome == 'success'


def test_generate_outcome(stub):
    task = Task(stub)
    assert task._generate_outcome(None, None, None).outcome == 'success'

    try:
        raise TypeError('mytest')
    except TypeError:
        exc_info = sys.exc_info()
    outcome = task._generate_outcome(*exc_info)
    assert outcome.outcome == 'failure'
    assert 'mytest' in outcome.message

    outcome = task._generate_outcome(TaskAborted, TaskAborted("a message"), None)
    assert outcome.outcome == 'failure'
    assert outcome.message == 'redis_tasks.exceptions.TaskAborted: a message\n'


class CMCheckMiddleware:
    """Check that no middleware functions are run inside the shutdown_cm"""
    def __init__(self):
        self.in_cm = False
        self.failed = False

    def __enter__(self):
        self.in_cm = True

    def __exit__(self, *args):
        self.in_cm = False

    def run_task(self, task, run, args, kwargs):
        self.failed |= self.in_cm
        try:
            run(*args, **kwargs)
        finally:
            self.failed |= self.in_cm

    def process_outcome(self, task, *exc_info):
        self.failed |= self.in_cm

    def __call__(self):
        return self


class _SpyMiddleware:
    history = []

    @classmethod
    def reset(cls):
        cls.history = []

    def run_task(self, task, run, args, kwargs):
        self.history.append((self, 'before', (task, *args)))
        if getattr(self, 'raise_before', None):
            raise self.raise_before
        try:
            run(*args)
        finally:
            self.history.append((self, 'after', (task, *args)))
            if getattr(self, 'raise_after', None):
                raise self.raise_after

    def process_outcome(self, *args):
        self.history.append((self, 'process_outcome', args))
        if getattr(self, 'outcome', None):
            if isinstance(self.outcome, Exception):
                raise self.outcome
            else:
                return self.outcome

    def __call__(self):
        return self


@pytest.fixture()
def SpyMiddleware():
    yield _SpyMiddleware
    _SpyMiddleware.reset()


def test_middleware_order(mocker, SpyMiddleware, stub):
    task = Task(stub)
    spies = [SpyMiddleware(), SpyMiddleware()]
    cmcheck = CMCheckMiddleware()
    mocker.patch('redis_tasks.task.task_middleware', new=[cmcheck, *spies])
    outcome = task.execute(shutdown_cm=cmcheck)
    assert outcome.outcome == "success"
    assert SpyMiddleware.history == [
        (spies[0], 'before', (task, )),
        (spies[1], 'before', (task, )),
        (spies[1], 'after', (task, )),
        (spies[0], 'after', (task, )),
        (spies[1], 'process_outcome', (task, None, None, None)),
        (spies[0], 'process_outcome', (task, None, None, None))]
    assert not cmcheck.failed


def test_middleware_raise_before(mocker, SpyMiddleware, stub):
    task = Task(stub)
    spies = [SpyMiddleware(), SpyMiddleware(), SpyMiddleware()]
    mocker.patch('redis_tasks.task.task_middleware', new=spies)
    spies[1].raise_before = ArithmeticError()
    outcome = task.execute()
    assert outcome.outcome == "failure"
    assert 'ArithmeticError' in outcome.message
    assert SpyMiddleware.history == [
        (spies[0], 'before', (task, )),
        (spies[1], 'before', (task, )),
        (spies[0], 'after', (task, )),
        (spies[2], 'process_outcome', (task, ArithmeticError, Something, Something)),
        (spies[1], 'process_outcome', (task, ArithmeticError, Something, Something)),
        (spies[0], 'process_outcome', (task, ArithmeticError, Something, Something))]


def test_middleware_raise_after(mocker, SpyMiddleware, stub):
    task = Task(stub)
    spies = [SpyMiddleware(), SpyMiddleware()]
    mocker.patch('redis_tasks.task.task_middleware', new=spies)
    spies[1].raise_after = ArithmeticError()
    outcome = task.execute()
    assert outcome.outcome == "failure"
    assert 'ArithmeticError' in outcome.message
    assert SpyMiddleware.history == [
        (spies[0], 'before', (task, )),
        (spies[1], 'before', (task, )),
        (spies[1], 'after', (task, )),
        (spies[0], 'after', (task, )),
        (spies[1], 'process_outcome', (task, ArithmeticError, Something, Something)),
        (spies[0], 'process_outcome', (task, ArithmeticError, Something, Something))]


def test_outcome_middlewares(mocker, SpyMiddleware, stub):
    task = Task(stub)
    spies = [SpyMiddleware() for i in range(2)]
    mocker.patch('redis_tasks.task.task_middleware', new=spies)
    assert task._generate_outcome(None, None, None).outcome == 'success'
    assert SpyMiddleware.history == [
        (spies[1], 'process_outcome', (task, None, None, None)),
        (spies[0], 'process_outcome', (task, None, None, None))]

    SpyMiddleware.reset()
    spies = [SpyMiddleware() for i in range(3)]
    mocker.patch('redis_tasks.task.task_middleware', new=spies)
    spies[2].outcome = True
    spies[1].outcome = ArithmeticError()
    spies[0].outcome = None
    sentinel = mocker.sentinel.error
    outcome = task._generate_outcome(None, sentinel, None)
    assert SpyMiddleware.history == [
        (spies[2], 'process_outcome', (task, None, sentinel, None)),
        (spies[1], 'process_outcome', (task, None, None, None)),
        (spies[0], 'process_outcome', (task, ArithmeticError, spies[1].outcome, Something))]
    assert outcome.outcome == 'failure'
    assert outcome.message.splitlines()[-1] == 'ArithmeticError'


def test_middleware_constructor_exception(SpyMiddleware, mocker, stub):
    task = Task(stub)
    spies = [SpyMiddleware() for i in range(2)]
    mws = [spies[0], "nope", spies[1]]
    mocker.patch('redis_tasks.task.task_middleware', new=mws)
    assert task.execute().outcome == 'failure'
    assert SpyMiddleware.history == [
        (spies[0], 'before', (task, )),
        (spies[0], 'after', (task, )),
        (spies[1], 'process_outcome', (task, TypeError, Something, Something)),
        (spies[0], 'process_outcome', (task, TypeError, Something, Something))]
    SpyMiddleware.reset()



import datetime
import random
import uuid

import pytest

from redis_tasks.exceptions import WorkerDoesNotExist
from redis_tasks.registries import failed_task_registry, worker_registry
from redis_tasks.task import TaskOutcome
from redis_tasks.utils import decode_list
from redis_tasks.worker import Worker, WorkerState
from tests.utils import QueueFactory, id_list


def test_state_transitions(time_mocker, connection, assert_atomic):
    worker = Worker('myworker', queues=[QueueFactory(), QueueFactory()])
    time = time_mocker('redis_tasks.worker.utcnow')
    queue = worker.queues[0]

    time.step()
    assert not connection.exists(worker.key, worker.task_key)
    assert worker.id not in worker_registry.get_worker_ids()
    with assert_atomic():
        worker.startup()
    assert worker.id in worker_registry.get_worker_ids()
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.started_at == time.now
        assert w.state == WorkerState.IDLE

    queue.enqueue_call()
    with assert_atomic(exceptions=['hgetall']):
        task = queue.dequeue(worker)
    assert connection.exists(worker.key, worker.task_key) == 2
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.current_task_id == task.id
        assert w.state == WorkerState.IDLE

    with assert_atomic():
        worker.start_task(task)
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.current_task_id == task.id
        assert w.state == WorkerState.BUSY

    with assert_atomic():
        worker.end_task(task, TaskOutcome("success"))
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.current_task_id is None
        assert w.state == WorkerState.IDLE

    time.step()
    assert connection.ttl(worker.key) == -1
    assert worker.id in worker_registry.get_worker_ids()
    with assert_atomic():
        worker.shutdown()
    assert worker.id not in worker_registry.get_worker_ids()
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.state == WorkerState.DEAD
        assert w.shutdown_at == time.now
    assert connection.ttl(worker.key) > 0


def test_died(time_mocker, connection, assert_atomic):
    time = time_mocker('redis_tasks.worker.utcnow')

    # Die while idle
    worker = Worker('idleworker', queues=[QueueFactory()])
    time.step()
    worker.startup()
    time.step()
    assert connection.ttl(worker.key) == -1
    assert worker.id in worker_registry.get_worker_ids()
    with assert_atomic():
        worker.died()
    assert worker.id not in worker_registry.get_worker_ids()
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.state == WorkerState.DEAD
        assert w.shutdown_at == time.now
    assert connection.ttl(worker.key) > 0

    # die whith task in limbo
    worker = Worker('limboworker', queues=[QueueFactory(), QueueFactory()])
    queue = worker.queues[1]
    time.step()
    worker.startup()
    time.step()
    queue.enqueue_call()
    task = queue.dequeue(worker)
    with assert_atomic(exceptions=['hgetall']):
        worker.died()
    assert queue.get_task_ids() == [task.id]
    assert worker.id not in worker_registry.get_worker_ids()
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.state == WorkerState.DEAD
        assert w.current_task_id is None
        assert w.shutdown_at == time.now
    assert connection.ttl(worker.key) > 0

    # die while busy
    worker = Worker('busyworker', queues=[QueueFactory(), QueueFactory()])
    queue = worker.queues[1]
    time.step()
    worker.startup()
    time.step()
    queue.enqueue_call()
    task = queue.dequeue(worker)
    worker.start_task(task)
    with assert_atomic(exceptions=['hgetall']):
        worker.died()
    assert queue.get_task_ids() == []
    assert failed_task_registry.get_task_ids() == [task.id]
    assert worker.id not in worker_registry.get_worker_ids()
    for w in [worker, Worker.fetch(worker.id)]:
        assert w.state == WorkerState.DEAD
        assert w.current_task_id is None
        assert w.shutdown_at == time.now
    assert connection.ttl(worker.key) > 0


def test_fetch_current_task():
    worker = Worker('testworker', queues=[QueueFactory()])
    queue = worker.queues[0]
    queue.enqueue_call()
    assert worker.fetch_current_task() is None
    task = queue.dequeue(worker)
    assert worker.fetch_current_task().id == task.id


def test_all():
    w1 = Worker('w1', queues=[QueueFactory()])
    w2 = Worker('w2', queues=[QueueFactory()])
    assert Worker.all() == []
    w1.startup()
    w2.startup()
    assert id_list(Worker.all()) == [w1.id, w2.id]
    w1.shutdown()
    assert id_list(Worker.all()) == [w2.id]


def test_heartbeat(mocker):
    heartbeat = mocker.patch.object(worker_registry, 'heartbeat')
    worker = Worker('testworker', queues=[QueueFactory()])
    worker.startup()
    worker.heartbeat()
    assert heartbeat.called_once_with(worker)


def test_persistence(assert_atomic, connection, time_mocker):
    time = time_mocker('redis_tasks.worker.utcnow')
    time.step()
    fields = {'description', 'state', 'queues', 'started_at', 'shutdown_at', 'current_task_id'}

    def randomize_data(worker):
        string_fields = ['description', 'state', 'current_task_id']
        date_fields = ['started_at', 'shutdown_at']
        for f in string_fields:
            setattr(worker, f, str(uuid.uuid4()))
        for f in date_fields:
            setattr(worker, f, datetime.datetime(
                random.randint(1000, 9999), 1, 1, tzinfo=datetime.timezone.utc))

        worker.args = tuple(str(uuid.uuid4()) for i in range(4))
        worker.kwargs = {str(uuid.uuid4()): ["d"]}
        worker.meta = {"x": [str(uuid.uuid4())]}
        worker.aborted_runs = ["foo", "bar", str(uuid.uuid4())]

    def as_dict(worker):
        return {f: getattr(worker, f) if f != 'queues' else [q.name for q in worker.queues]
                for f in fields}

    worker = Worker("testworker", queues=[QueueFactory()])
    with assert_atomic():
        worker.startup()
    assert as_dict(worker) == as_dict(Worker.fetch(worker.id))
    worker2 = Worker("worker2", queues=[QueueFactory() for i in range(5)])
    worker2.startup()
    assert as_dict(worker2) == as_dict(Worker.fetch(worker2.id))
    assert as_dict(worker) != as_dict(worker2)

    worker = Worker("testworker", queues=[QueueFactory()])
    worker.startup()
    with assert_atomic():
        worker._save()
    assert set(decode_list(connection.hkeys(worker.key))) <= fields
    assert as_dict(Worker.fetch(worker.id)) == as_dict(worker)

    randomize_data(worker)
    worker._save()
    assert as_dict(Worker.fetch(worker.id)) == as_dict(worker)

    # only deletes
    worker.started_at = None
    worker._save(['started_at'])
    assert as_dict(Worker.fetch(worker.id)) == as_dict(worker)

    for i in range(5):
        store = random.sample(fields, 3)
        copy = Worker.fetch(worker.id)
        randomize_data(copy)
        copy._save(store)
        for f in store:
            setattr(worker, f, getattr(copy, f))
        assert as_dict(Worker.fetch(worker.id)) == as_dict(worker)

    worker = Worker("nonexist", queues=[QueueFactory()])
    with pytest.raises(WorkerDoesNotExist):
        worker.refresh()
    with pytest.raises(WorkerDoesNotExist):
        Worker.fetch("nonexist")



import datetime
import logging
import random
import signal
import threading
import time
import uuid
from contextlib import suppress

import croniter

from .conf import connection, construct_redis_key, settings
from .exceptions import TaskDoesNotExist
from .queue import Queue
from .smear_dst import DstSmearingTz
from .task import Task, TaskStatus
from .utils import (
    LazyObject, atomic_pipeline, decode_dict, utcformat, utcnow, utcparse)

logger = logging.getLogger(__name__)

local_tz = LazyObject(lambda: DstSmearingTz(settings.TIMEZONE))


class CrontabSchedule:
    def __init__(self, crontab):
        self.crontab = crontab

    def get_next(self, after):
        after = local_tz.from_utc(after)
        iter = croniter.croniter(self.crontab, after, ret_type=datetime.datetime)
        return local_tz.to_utc(iter.get_next())


crontab = CrontabSchedule


def once_per_day(time_str):
    hour, minute = time_str.split(':')
    return CrontabSchedule(f'{minute} {hour} * * *')


class PeriodicSchedule:
    def __init__(self, *, hours=0, minutes=0, seconds=0, start_at=None):
        self.interval = hours * 60 * 60 + minutes * 60 + seconds
        if start_at is None:
            start_at = random.uniform(0, self.interval)
        elif isinstance(start_at, str):
            hours, minutes = start_at.split(':')
            start_at = int(hours) * 60 * 60 + int(minutes) * 60
        self.start_at = start_at
        assert self.interval < 12 * 60 * 60
        assert self.start_at < 24 * 60 * 60

    def get_next(self, after):
        after = after.astimezone(local_tz.tz)
        midnight = local_tz.tz.localize(
            after.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None),
            is_dst=True)
        after_time = (after - midnight).total_seconds()
        if self.start_at > after_time:
            next_time = self.start_at
        else:
            since_start = after_time - self.start_at
            remaining = self.interval - since_start % self.interval
            next_time = after_time + remaining

        next = local_tz.tz.normalize(midnight + datetime.timedelta(seconds=next_time))
        if next.day != after.day:
            midnight = local_tz.tz.localize(
                after.replace(hour=23, minute=59, second=59, microsecond=0, tzinfo=None),
                is_dst=False) + datetime.timedelta(seconds=1)
            next = midnight + datetime.timedelta(seconds=self.start_at)
        return next.astimezone(datetime.timezone.utc)


run_every = PeriodicSchedule


class SchedulerEntry:
    def __init__(self, id, config):
        self.id = id
        self.key = construct_redis_key(f"schedule_entry:{self.id}")
        self.singleton = config.get('singleton', True)
        self.task_template = [config['task'],
                              config.get('args', ()), config.get('kwargs', {})]
        self.queue = Queue(config.get('queue', settings.SCHEDULER_QUEUE))
        self.schedule = config['schedule']

        self.last_save = None
        stored = decode_dict(connection.hgetall(self.key))
        prev_run = stored.get("prev_run")
        self.prev_run = utcparse(prev_run) if prev_run else utcnow()
        self.prev_task_id = stored.get("prev_task_id")

        self.next_run = self.schedule.get_next(self.prev_run)

        # Make sure task config is valid
        Task(*self.task_template)

    @atomic_pipeline
    def save(self, *, pipeline):
        pipeline.hset(self.key, "prev_run", utcformat(self.prev_run))
        if self.prev_task_id:
            pipeline.hset(self.key, "prev_task_id", self.prev_task_id)
        else:
            pipeline.hdel(self.key, "prev_task_id")
        ttl = max(24 * 60 * 60, settings.SCHEDULER_MAX_CATCHUP * 5)
        pipeline.expire(self.key, ttl)
        self.last_save = utcnow()

    @atomic_pipeline
    def process(self, now, *, pipeline):
        max_catchup = now - datetime.timedelta(seconds=settings.SCHEDULER_MAX_CATCHUP)
        self.next_run = self.schedule.get_next(max(max_catchup, self.prev_run))
        if self.next_run > now:
            if (not self.last_save or
                    (now - self.last_save).total_seconds() >= settings.SCHEDULER_MAX_CATCHUP):
                self.save(pipeline=pipeline)
            return

        self.prev_run = now

        if self.singleton:
            self.next_run = self.schedule.get_next(now)
            if self.is_enqueued():
                logger.info(f'Schedule entry "{self.id}" already enqueued or running, skipping')
            else:
                self.enqueue(pipeline=pipeline)
        else:
            while self.next_run <= now:
                self.enqueue(pipeline=pipeline)
                self.next_run = self.schedule.get_next(self.next_run)
        self.save(pipeline=pipeline)

    def is_enqueued(self):
        if self.prev_task_id:
            with suppress(TaskDoesNotExist):
                prev_task = Task.fetch(self.prev_task_id)
                if prev_task.status not in [TaskStatus.FINISHED, TaskStatus.FAILED]:
                    return True
        return False

    @atomic_pipeline
    def enqueue(self, *, pipeline):
        task = self.queue.enqueue_call(*self.task_template, pipeline=pipeline)
        self.prev_task_id = task.id
        return task


class Scheduler:
    def __init__(self):
        self.schedule = [SchedulerEntry(k, v) for k, v in settings.SCHEDULE.items()]
        self.shutdown_requested = threading.Event()

    def setup_signal_handler(self):
        def stop(signum, frame):
            logger.info('Initiating redis_tasks scheduler shutdown')
            self.shutdown_requested.set()

        signal.signal(signal.SIGINT, stop)
        signal.signal(signal.SIGTERM, stop)

    def run(self):
        if not self.schedule:
            logger.error("No schedule configured, nothing to do")
            return

        self.setup_signal_handler()

        HEARTBEAT_FREQ = 10
        with Mutex(timeout=HEARTBEAT_FREQ + 2) as mutex:
            logger.info('redis_tasks scheduler started')
            while not self.shutdown_requested.is_set():
                mutex.extend()
                now = utcnow()
                for entry in self.schedule:
                    entry.process(now)

                next_run = min(x.next_run for x in self.schedule)
                next_heartbeat = now + datetime.timedelta(seconds=HEARTBEAT_FREQ)
                wait_for = (min(next_run, next_heartbeat) - utcnow()).total_seconds()
                self.shutdown_requested.wait(wait_for)
        logger.info('redis_tasks scheduler shut down')


def scheduler_main():
    Scheduler().run()


class Mutex(object):
    expire_script = None

    def __init__(self, *, timeout):
        self.key = construct_redis_key('scheduler')
        self.timeout = timeout
        self.token = None

        # KEYS[1]: lock key, ARGS[1]: token, ARGS[2]: milliseconds
        # return 1 if the lock was held and the expire executed, otherwise 0
        self.expire_script = connection.register_script("""
            if redis.call('get', KEYS[1]) == ARGV[1] then
                return redis.call('pexpire', KEYS[1], ARGV[2])
            else
                return 0
            end""")

    def __enter__(self):
        if not self.acquire(wait=False):
            wait_for = self.timeout + 1
            logger.warning("Found signs of an already running scheduler instance, "
                           f"waiting {wait_for} seconds for it to disappear")
            try:
                self.acquire(wait=wait_for)
            except TimeoutError:
                raise RuntimeError("redis_tasks scheduler already running")
        return self

    def __exit__(self, *exc):
        if self.token:
            self.expire_script(keys=[self.key], args=[self.token, 0])
            self.token = None

    def acquire(self, wait=None):
        token = str(uuid.uuid1()).encode()
        stop_trying_at = time.time() + wait
        while True:
            acquired = connection.set(self.key, token, nx=True, px=int(self.timeout * 1000))
            if acquired:
                self.token = token
                return True
            elif not wait:
                return False
            elif time.time() > stop_trying_at:
                raise TimeoutError
            time.sleep(0.1)

    def extend(self):
        if not self.expire_script(keys=[self.key], args=[self.token, int(self.timeout * 1000)]):
            raise RuntimeError("Cannot refresh a lock that's no longer owned")



import os

from setuptools import find_packages, setup


def get_version():
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, 'redis_tasks/__init__.py')) as f:
        version_line = next(line for line in f
                            if line.startswith('__version__'))
        return eval(version_line.split('=')[1])
    raise RuntimeError('No version info found.')


def get_long_description():
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


setup(
    name='redis-tasks',
    version=get_version(),
    url='https://github.com/djangsters/redis-tasks/',
    author='Christian Fersch',
    author_email='christian@djangsters.de',
    description='redis-tasks is a lightweight library for processing background tasks',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    python_requires='>=3.7',
    install_requires=['redis >=3.2.0',
                      'click',
                      'croniter >=0.3.23',
                      'pytz'],
    entry_points={
        'console_scripts': [
            'redis_tasks = redis_tasks.cli:main',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ]
)



# flake8: noqa
from .queue import Queue
from .worker_process import PostponeShutdown, worker_main, TWorker
from .task import redis_task, get_current_task, Task
from .exceptions import *
from .scheduler import crontab, scheduler_main, once_per_day, run_every

__version__ = '0.0.11'



import os
from types import SimpleNamespace

from redis_tasks import conf, defaults


def test_settings(mocker):
    mocker.patch.dict(os.environ, {conf.ENVIRONMENT_VARIABLE: 'tests.app.settings_a'})
    s = conf.Settings()
    assert s.REDIS_PREFIX == "test_foo"
    assert s.EXTRA_OPTION == "bar"
    assert s.WORKER_HEARTBEAT_FREQ == defaults.WORKER_HEARTBEAT_FREQ

    s = conf.Settings()
    s.configure(SimpleNamespace(REDIS_PREFIX="barbar"))
    assert s.REDIS_PREFIX == "barbar"
    assert s.WORKER_HEARTBEAT_FREQ == defaults.WORKER_HEARTBEAT_FREQ


def test_mock_settings(settings):
    assert settings.DEFAULT_TASK_TIMEOUT == defaults.DEFAULT_TASK_TIMEOUT
    settings.DEFAULT_TASK_TIMEOUT = "foo"
    assert conf.settings.DEFAULT_TASK_TIMEOUT == "foo"


def test_mock_settings_after(settings):
    assert conf.settings.DEFAULT_TASK_TIMEOUT != "foo"
    assert settings.DEFAULT_TASK_TIMEOUT != "foo"


def test_construct_redis_key(settings):
    settings.REDIS_PREFIX = "bar"
    assert conf.construct_redis_key("foo") == "bar:foo"
